//This agent can offer and accept 
public class Agent {
	int id,
		maxOffer,
		minAccept,
		sumToDivide,
		maxRounds,
		maxToTradeWith,
		currentStrategy;
	float myScore;
	int [] informationBoughtHigh,
			informationBoughtLow;
	int [][]	offers,
				received;
	boolean [][] myBidsAccepted,
				 bidsIAccepted;
	boolean debug;
	public static final int OPTIMUM_BID = 0;
	public static final int RANDOM_BID = 1;
	public static final int FIXED_BID = 2;

	public Agent(int id, int maxOffer, int minAccept, int maxToTradeWith, int sumToDivide, int maxRounds, boolean debug)
	{
		this.id = id;
		this.maxOffer = maxOffer;
		this.minAccept = minAccept;
		this.sumToDivide = sumToDivide;
		this.maxRounds = maxRounds;
		this.maxToTradeWith = maxToTradeWith;
		this.debug = debug;
		myScore = 0;
		offers = new int[maxToTradeWith][maxRounds];
		received = new int[maxToTradeWith][maxRounds];
		myBidsAccepted = new boolean[maxToTradeWith][maxRounds];
		bidsIAccepted = new boolean[maxToTradeWith][maxRounds];
		informationBoughtHigh = new int[maxToTradeWith];
		informationBoughtLow = new int[maxToTradeWith];
		initialize(informationBoughtHigh);
		initialize(informationBoughtLow);
		initialize(offers);
		initialize(received);
	}
	
	public void initialize(int [][] a)
	{
		for (int i = 0; i < a.length; i++)
		{
			for (int j = 0; j < a[0].length; j++)
			{
				a[i][j] = -1;
			}
		}
	}
	
	public void initialize(float [][] a)
	{
		for (int i = 0; i < a.length; i++)
		{
			for (int j = 0; j < a[0].length; j++)
			{
				a[i][j] = -1;
			}
		}
	}
	
	public void initialize(int []a)
	{
		for (int i = 0; i < a.length; i++)
		{
				a[i] = -1;
		}
	}
	
	public void initialize(float []a)
	{
		for (int i = 0; i < a.length; i++)
		{
				a[i] = -1;
		}
	}
	
	//There is a neater way to do this, but this function determines
	//whether this agent will always bid it's optimum or offers it's maximum
	//bid.
	public void setBiddingStrategy(String strategy)
	{
		if (strategy.contentEquals("OptimumBid"))
			currentStrategy = OPTIMUM_BID;
		else if (strategy.contentEquals("FixedBid"))
			currentStrategy = FIXED_BID;
		else if (strategy.contentEquals("RandomBid"))
			currentStrategy = RANDOM_BID;
		else
		{
			System.err.print("Incorrect Strategy given, please retry");
			System.exit(0);
		}
	}
	
	public void printOffers( int thisRecipient,int thisRound)
	{	
		for (int r = 0; r <= thisRound; r++)
		{
			System.out.print("Round " + r + "\t");
		}
		System.out.println("");
		for (int recipient = 0; recipient < maxToTradeWith; recipient++)
		{
			for (int round = 0; round <= thisRound; round++)
			{
				if (round != thisRound)
				{
					if (myBidsAccepted[recipient][round])
						System.out.print("V " + offers[recipient][round] + "\t");
					else
						System.out.print("X " + offers[recipient][round] + "\t");
				}
				
				else if (recipient <= thisRecipient)
				{
					if (myBidsAccepted[recipient][round])
						System.out.print("V " + offers[recipient][round] + "\t");
					else
						System.out.print("X " + offers[recipient][round] + "\t");
				}				
			}
			System.out.println("\n");
		}
	}
	
	public int minAccept()
	{
		return minAccept;
	}
	
	public void printReceived()
	{
		System.out.println("OffersReceived by agent " + id + " minAccept: " + minAccept);
		
		for (int r = 0; r < maxRounds; r++)
		{
			System.out.print("Round " + r + "\t");
		}
		
		System.out.println("");
		
		for (int bidder = 0; bidder < received.length; bidder++)
		{
			for (int round = 0; round < maxRounds; round++)
			{
				if (bidsIAccepted[bidder][round])
					System.out.print("V " + received[bidder][round] + "\t");
				else
					System.out.print("X " + received[bidder][round] + "\t");
			}
			System.out.println("\n");
		}
	}
	
	//Identification of an agent
	public int askId()
	{
		return id;
	}
	
	public void pay(float amount)
	{
		myScore -= amount;
	}
	
	//This agent will bid on information depending on the expected extra profit
	//this information will give
	public float auctionBid(Agent r, int infoAccLow, int infoAccHigh)
	{
		System.out.println("-------------------------------------");
		//System.out.println("Old table of acceptance");
		//First determine the bid that the agent would have done prior to the new info
		int [][] acceptedTable = calculateAcceptedTable(maxOffer, acceptLow(r, true), acceptHigh(r, true));
		int [] totals = calculateTotals(acceptedTable, infoAccLow);
		int priorBid = idxHighest(totals);
		 
		System.out.println("priorBid: " + priorBid + ", maxOffer: " + maxOffer);
		System.out.println("prior acceptLow: " + acceptLow(r, true) + ", prior acceptHigh: " + acceptHigh(r, true));

		//Determine new Ranges for the recipient, using both the new info and the already
		//acquired information.
		int newAcceptLow = max(acceptLow(r, true), infoAccLow);
		int newAcceptHigh = min(infoAccHigh, acceptHigh(r, true));
		
		System.out.println("post acceptLow: " + newAcceptLow + ", post acceptHigh: " + newAcceptHigh);
		
		//Use the new Ranges to compute a new AcceptedTable and a new best bid
		//System.out.println("New table of acceptance");
		
		//!!Do not use newAccHigh here, or old best offer could be evaluated as zero
		int [][] newAcceptedTable = calculateAcceptedTable(maxOffer, newAcceptLow, newAcceptHigh);
		int [] newTotals = calculateTotals(newAcceptedTable, newAcceptLow);
		int postBid = idxHighest(newTotals);
		
		int range;
		//example high: 2, low: 0: range = 2 
		if (newAcceptLow == 0)
			range = (newAcceptHigh - newAcceptLow );
		else //example high: 2, low: 2: range = 1 or high 4, low 2 range:3
			range = (newAcceptHigh - newAcceptLow + 1);
		
		float evPrior;
		//Use the new information to calculate the expected prior profit
		if (priorBid >= newTotals.length)
		{
			evPrior = 0;
			System.out.println("priorBid >= newTotals.length");
			System.exit(0);
		}
		else
			evPrior = newTotals[priorBid] / (float)range;	
				
		float evPost =  newTotals[postBid] / (float)(range);
		
		//small check to eliminate negative values for the ev's
		//maxEvPost = (maxEvPost < 0) ? 0 : maxEvPost;
		//maxEvPrior = (maxEvPrior < 0) ? 0: maxEvPrior;
		
		System.out.println("evPrior: " + evPrior);
		System.out.println("evPost: " + evPost);
		
		//bidding is done with difference
		return evPost - evPrior;
	}
	
	public void storeInformation(int low, int high, int agent)
	{
		informationBoughtLow[agent] = low;
		informationBoughtHigh[agent] = high;
	}
	
	//	returns the highest offer that is known to be rejected by the recipient
	public int acceptLow(Agent r, boolean useBoughtInformation)
	{
		int minBid = 0;
		//check for earlier offers made to this agent
		for (int prevRound = 0; prevRound < maxRounds; prevRound++)
		{
			//If the bid was rejected
			//And it was the highest bid that was rejected, set this to the minBid 
			if (myBidsAccepted[r.askId()][prevRound]==false && offers[r.askId()][prevRound] >= minBid)
				minBid = offers[r.askId()][prevRound] + 1;
		}
		//	no extra information was bought about this agent or we don't want to use that info
		if (informationBoughtLow[r.askId()] == -1 || !useBoughtInformation)
			return minBid;
		else // there was information bought
			return max(minBid, informationBoughtLow[r.askId()]);
	}
	
	//	returns the lowest offer that is known to be accepted by the recipient
	public int acceptHigh(Agent r, boolean useBoughtInformation)
	{
		//this value is initialized to 
		int lowestAcc = 10;
		//check for earlier offers made to this agent
		for (int prevRound = 0; prevRound < maxRounds; prevRound++)
		{
			//If the last bid was accepted
			//And it was the lowest bid that was accepted, set this to the maxBid
			if (myBidsAccepted[r.askId()][prevRound]==true && offers[r.askId()][prevRound] < lowestAcc)
				lowestAcc = offers[r.askId()][prevRound];
		}
		// no extra information was bought about this agent or we don't want to use that info
		if (informationBoughtHigh[r.askId()] == -1 || !useBoughtInformation)
			return lowestAcc;
		else // there was information bought
			return min(lowestAcc, informationBoughtHigh[r.askId()]);

	}
	
	//Returns the index of the highest number in the array
	private int idxHighest(int [] array)
	{
		int highest = -Integer.MAX_VALUE;
		int idx = -1;
		for (int i = 0; i < array.length; i++)
		{
			if (array[i] > highest)
			{
				highest = array[i];
				idx = i;
			}
		}
		return idx;
	}
	
	private int min(int a, int b)
	{
		if ( a < b)
			return a;
		else return b;
	}
	
	private int max(int a, int b)
	{
		if ( a > b)
			return a;
		else return b;
	}

	
	public int[][] calculateAcceptedTable(int maxOffer, int accLow, int accHigh)
	{
		//The range for the table is limited by the thresholds the agent
		//still believes the recipient can still have, and by the maximum an agent
		//can offer. There is no use in considering agents that only accept more
		//than this agents maxOffer, since they will always refuse the offer.
		//
		//int range = min(accHigh, maxOffer) + 1;
		int range = maxOffer + 1;
		
		int tempRange = accHigh+ 1;
		
		int [][] acceptedTable = new int[range][tempRange];
		
		//if the agent can't bid high enough return uninitialized array
		if (accLow > maxOffer)
			return acceptedTable;	
		
		//Do for each threshold still hold possible for this recipient
		for (int xAcc = max(accLow,1); xAcc <= accHigh; xAcc++)
		{			
			//For each offer this agent can make, but no higher than the minimum
			//the recipient is willing to accept
			for (int offer = max(accLow,1); offer <= maxOffer; offer++)
			{
				if (offer >= xAcc) 	//accept
					acceptedTable[offer][xAcc] = (sumToDivide - offer);
				else //refuse
					acceptedTable[offer][xAcc] = 0;
			}
		}
		
		if (debug)
			printAcceptedTable(acceptedTable, maxOffer, accLow, accHigh);
		
		return acceptedTable;
	}
	
	private void printAcceptedTable(int[][] acceptedTable, int maxOffer, int accLow, int accHigh)
	{		
		System.out.println("Table of acceptance: ");
		System.out.print("Bid: \t");
		
		for (int i = max(accLow,1); i < acceptedTable.length; i++)
		{
				System.out.print(i + "\t");
		}
		System.out.print("\n");
		
		//print a line
		for (int i = max(accLow,1); i <= acceptedTable.length; i++)
		{
			System.out.print("--------");
		}
		System.out.println();
		
		//Do for the limits of the Recipient the agent still holds possible
		for (int xAcc = max(accLow,1); xAcc <= min(accHigh, maxOffer); xAcc++)
		{			
			//For each offer this agent can make
			System.out.print("Xacc:" + xAcc + "\t");
			for (int offer = max(accLow,1); offer <= maxOffer; offer++)
			{
				System.out.print(acceptedTable[offer][xAcc] + "\t");
			}
			System.out.println();
		}
		
		//print a line
		for (int i = max(accLow,1); i <= acceptedTable.length; i++)
		{
			System.out.print("--------");
		}
		
		System.out.println();
	}
	
	public int[] calculateTotals(int acceptedTable[][], int accLow)
	{
		int [] totals = new int[acceptedTable.length];
		
		for (int offer = 0; offer < acceptedTable.length; offer++)
		{
			for (int xAcc = 0; xAcc < acceptedTable[0].length; xAcc++)
			{
				totals[offer] += acceptedTable[offer][xAcc];
			}
		}
		if (debug)
			printTotals(totals, accLow);
		return totals;
	}
	
	private void printTotals(int [] totals, int accLow)
	{
		System.out.print("Total:\t");
		for (int i = max(accLow,1); i < totals.length; i++)
		{
			System.out.print(totals[i] + "\t");
		}
		System.out.println();
	}
	//This agent determines an optimum bid, based on the fact that all of the possible thresholds
	//the recipient could have are equally likely.
	//minAccept is the minimum offer the recipient is willing to accept
	//maxAccept is an offer the recipient is known to accept
	private int determineOptimumBid(int maxOffer, int accLow, int accHigh)
	{
		int [][] acceptedTable = calculateAcceptedTable(maxOffer, accLow, accHigh);
		int [] totals = calculateTotals(acceptedTable, accLow);
		int bestBid = idxHighest(totals);
		return bestBid;
	}	
	
	public void doBid(int offer, Agent r, int round)
	{
		//remember the offer you offered this agent
		offers[r.askId()][round] = offer;
		
		//this is the actual bidding!
		if (r.evaluateOffer(id, offer, round))
		{
			myBidsAccepted[r.askId()][round] = true;
			myScore += sumToDivide - offer;
		}
		
		if (debug)
		{
			System.out.println( "Bid from agent " + id + "(maxOffer: " + maxOffer +")" + " to recipient " + r.askId()
					+ " [low: " + acceptLow(r, true) + ", high:" + acceptHigh(r, true) + "]\noffering: " + offer +
			", myScore: " + myScore);
		}
	}
	
	public int determineBid(Agent r, int round)
	{
		//check this recipient for the known ranges
		int accLow = acceptLow(r, true);
		int accHigh = acceptHigh(r, true);

		//if the agent can't bid high enough bid -1
		int offer = -1;
		
		if (accLow <= maxOffer) {
			switch (currentStrategy) {
			case OPTIMUM_BID: 
				offer = determineOptimumBid(maxOffer, accLow, accHigh);
				break;
			case RANDOM_BID: 
				offer = (int)(Math.random() * (accHigh - accLow)) + accLow;
				break;
			case FIXED_BID:
				offer = maxOffer;
				break;
			}
		}
		return offer;
	}
	
	public boolean evaluateOffer(int id, int offer, int round)
	{
		//recieved an offer
		received[id][round] = offer;
		
		//some decision things here
		if (offer >= minAccept)
		{
			bidsIAccepted[id][round] = true;
			myScore += offer;
			return true;
		}
		else
			return false;
	}
	
	public float myScore()
	{
		return myScore;
	}
}